/* File: $Id: bus.support.cc,v 1.19 1998/02/25 16:17:56 morsiani Exp $ */

/****************************************************************************
 * 
 * This module implements some classes used from the SystemBus for the
 * scheduling of device events (such as device operations completion and
 * interrupts generation).  They are: Event class, to keep track of single
 * events, as required by devices; and EventQueue class, to organize the
 * Events into a time-ordered queue.
 *
 ***************************************************************************/


/***************************************************************************/
/* Inclusion of header files.                                              */
/***************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "h/const.h"
#include "h/types.h"


/****************************************************************************/
/* Inclusion of imported declarations.                                      */
/****************************************************************************/

#include "e/utility.e"


/****************************************************************************/
/* Declarations strictly local to the module.                               */
/****************************************************************************/


/****************************************************************************/
/* Definitions to be exported.                                              */
/****************************************************************************/

// Event class is used to keep track of the external events of the
// system: device operations and interrupt generation.
// Every object contains a device number saying what device will complete
// its operation and generate a interrupt, a TimeStamp saying when it 
// will happen, and a pointer to the next event in a structure, if needed
// (set to NULL otherwise) 

class Event
{
	public: 
		
		// This method creates a new Event object and initalizes it
		Event(TimeStamp * ts, Word inc, unsigned int il, unsigned int dev);
		
		// This method deletes a Event object and its internal structures
		~Event(void);
		
		// This method returns the interrupt line of the device requiring
		// the Event
		unsigned int getIntLine(void);
		
		// This method returns the device number of the device requiring
		// the Event
		unsigned int getDevNum(void);

		// This method returns the TimeStamp access pointer
		TimeStamp * getTS(void);
		
		// This method links an Event to its successor in a structure
		void AddBefore(Event * ev);
		
		// This method inserts an Event after another, linking the former to the
		// successor of the latter
		void InsAfter(Event * ev);
		
		// This method returns the pointer to the successor of an Event
		Event * Next(void);
	
	private:

		// interrupt line and device number identifying the device which 
		// requires the Event
		unsigned int intL;
		unsigned int devNum;

		// Event verification time
		TimeStamp * time;

		Event * next;
};


// This method creates a new Event object and initalizes it
Event::Event(TimeStamp * ts, Word inc, unsigned int il, unsigned int dev)
{
	time = new TimeStamp(ts, inc);
	intL = il;
	devNum = dev;
	next = NULL;
}


// This method deletes a Event object and its internal structures
Event::~Event(void)
{
	delete time;
}



// This method returns the interrupt line of the device requiring the Event
unsigned int Event::getIntLine(void)
{
	return(intL);
}


// This method returns the device number of the device requiring the Event
unsigned int Event::getDevNum(void)
{
	return(devNum);
}


// This method returns the TimeStamp access pointer
TimeStamp * Event::getTS(void)
{
	return (time);
}


// This method links an Event to its successor in a structure
void Event::AddBefore(Event * ev)
{
	next = ev;
}


// This method inserts an Event after another, linking the former to the 
// successor of the latter
void Event::InsAfter(Event * ev)
{
	next = ev->next;
	ev->next = this;
}


// This method returns the pointer to the successor of an Event
Event * Event::Next(void)
{
	return (next);
}


/****************************************************************************/


// This class implements a NULL-terminated sorted queue of Event objects,
// used to schedule the device events in the system.
// An object contains a pointer to the head of the queue, and a pointer to
// the last inserted Event (to speed up insertion).
// The queue is sorted on ascending timestamp order

class EventQueue
{
	public:
		
		// This method creates a new (empty) queue
		EventQueue(void);
		
		// This method deletes the queue and its associated structures
		~EventQueue(void);
		
		// This method returns TRUE if the queue is empty, FALSE otherwise
		Boolean IsEmptyQ(void);
		
		// This method returns the TimeStamp in the EventQueue
		// head, if queue is not empty, and NULL otherwise
		TimeStamp * getHTS(void);

		// This method returns the interrupt line of the EventQueue head, if
		// not empty (0 otherwise)
		unsigned int getHIntLine(void);
		
		// This method returns the device number of the EventQueue head, if
		// not empty (0 otherwise)
		unsigned int getHDevNum(void);

		// This method creates a new Event object and inserts it in the
		// EventQueue; EventQueue is sorted on ascending time order
		TimeStamp * InsertQ(TimeStamp * ts, Word inc, unsigned int il, unsigned int dev);
		
		// This method removes the head of a (not empty) queue and sets it to the
		// following Event		
		void RemoveHead(void);
		
	private:
		
		// head of the queue
		Event * head;
		
		// last Event inserted
		Event * lastIns;
};

 
// This method creates a new (empty) queue
EventQueue::EventQueue(void)
{
	head = NULL;
	lastIns = NULL;
}


// This method deletes the queue and its associated structures
EventQueue::~EventQueue(void)
{
	Event *p, *q;
	
	p = head;
	q = NULL;
	while (p != NULL)
	{
		q = p->Next();
		delete p;
		p = q;
	}
}


// This method returns TRUE if the queue is empty, FALSE otherwise
Boolean EventQueue::IsEmptyQ(void)
{
	if (head == NULL)
		return(TRUE);
	else
		return(FALSE);
}


// This method returns a pointer to the timestamp of the EventQueue head, 
// if queue is not empty, and NULL otherwise
TimeStamp * EventQueue::getHTS(void)
{
	if (!IsEmptyQ())
		return(head->getTS());
	else 
		return(NULL);
}


// This method returns the interrupt line of the EventQueue head, 
// if not empty (0 otherwise)
unsigned int EventQueue::getHIntLine(void)
{
	if (!IsEmptyQ())
		return(head->getIntLine());
	else 
		return(0);
}


// This method returns the device number of the EventQueue head, 
// if not empty (0 otherwise)
unsigned int EventQueue::getHDevNum(void)
{
	if (!IsEmptyQ())
		return(head->getDevNum());
	else 
		return(0);
}


// This method creates a new Event object and inserts it in the
// EventQueue; EventQueue is sorted on ascending time order
TimeStamp * EventQueue::InsertQ(TimeStamp * ts, Word inc, unsigned int il, unsigned int dev)
{
	Event *ins, *p, *q;
	
	ins = new Event(ts, inc, il, dev);
	if (IsEmptyQ())
		head = ins;
	else 
		if ((ins->getTS())->LessEq(head->getTS()))
		{
			// "ins" has to happen before that at the head of the queue;
			// should be put before it
			ins->AddBefore(head);
			head = ins;
		}
		else
		{
			// should find place in queue: check lastIns to shorten search time 
			if (lastIns != NULL && !((ins->getTS())->LessEq(lastIns->getTS())))
				// can start from lastIns
				p = lastIns;
			else
				// must start from the head
				p = head;

			q = p;
			while (p != NULL && (p->getTS())->LessEq(ins->getTS()))
			{	
				// traverse the queue 
				q = p;
				p = p->Next();
			}
			// place found: insert after q and before p
			ins->InsAfter(q);
		}			
	lastIns = ins;	
	return(ins->getTS());
}


// This method removes the head of a (not empty) queue and sets it to the
// following Event
void EventQueue::RemoveHead(void)
{
	Event * p;
	
	if (!IsEmptyQ())
	{
		p = head;
		head = head->Next();

		if (p == lastIns)
			// reposition lastIns to the new head  
			lastIns = head;	

		delete p;
	}
}


/****************************************************************************/
/* Definitions strictly local to the module.                                */
/****************************************************************************/

