/* File: $Id: memspace.cc,v 1.20 2004/03/14 10:41:55 morsiani Exp morsiani $ */

/****************************************************************************
 *
 * This module implements the classes which represents memory spaces:
 * RamSpace (for RAM) and BiosSpace (for ROMs), completely under control of 
 * SystemBus.
 * 
 ****************************************************************************/


/****************************************************************************/
/* Inclusion of header files.                                               */
/****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "h/const.h"
#include "h/types.h"
#include "h/blockdev.h"
#include "h/aout.h"


/****************************************************************************/
/* Inclusion of imported declarations.                                      */
/****************************************************************************/

#include "e/utility.e"


/****************************************************************************/
/* Declarations strictly local to the module.                               */
/****************************************************************************/

HIDDEN Word * readBIOSFile(const char * fName, Word * sizep);
HIDDEN Word * emptyFrameSignal(const char * fName, Word * sizep);


/****************************************************************************/
/* Definitions to be exported.                                              */
/****************************************************************************/


// This class implements the RAM device. Any object allows reads and writes
// with random access to word-sized items using appropriate methods.
// Contents may be loaded from file at creation.  SystemBus must do all
// bounds checking and address conversion for access

class RamSpace 
{
	public:
	
		// This method creates a RamSpace object of a given size (in words)
		// and fills it with file contents if needed
		RamSpace(Word siz, const char * fName);
		
		// This method deletes the contents of a RamSpace object
		~RamSpace(void);
		
		// This method returns the value of Word at ofs address
		// (SystemBus must assure that ofs is in range)
		Word MemRead(Word ofs);
		
		// This method allows to write data to a specified address (as word
		// offset). SystemBus must check address validity and make
		// byte-to-word address conversion)
		void MemWrite(Word ofs, Word data);

		// This method returns RamSpace size in bytes
		Word Size(void);
		
	private:
		// pointer to memory structure
		Word * memPtr;
		
		// size of structure in words (C style addressing: [0..size - 1])
		Word size;
};


// This method creates a RamSpace object of a given size (in words) and
// fills it with core file contents if needed
RamSpace::RamSpace(Word siz, const char * fName)
{
	FILE * cFile = NULL;
	Word tag;
	unsigned int i;
	
	size = siz;
	memPtr = new Word [siz];
	
	if (fName != NULL && !SAMESTRING(fName, EMPTYSTR))
	{
		// tries to load core file from disk
		if ((cFile = fopen(fName, "r")) == NULL || \
			fread((void *) &tag, WORDLEN, 1, cFile) != 1 || \
			tag != COREFILEID)
		{
			ShowAlert("Unable to load core file", fName, "RAM will be filled with (nop)");
			for (i = 0; i < size; i++)
				memPtr[i] = NOP;
		}
		else
		{	
			fread((void *) memPtr, WORDLEN, size, cFile);
			if (!feof(cFile))
			{
				// core too large: it did not fit in memory
				ShowAlert("Core file did not fit in memory: too large", \
					"Increase setup RAM size and reset the system", EMPTYSTR);
			}
		}
		if (cFile != NULL)
			fclose(cFile);
	}
	// else no core has to be loaded
}


// This method deletes the contents of a RamSpace object
RamSpace::~RamSpace(void)
{
	delete memPtr;
}


// This method returns the value of Word at ofs address
// (SystemBus must assure that ofs is in range)
Word RamSpace::MemRead(Word ofs)
{
	if (ofs < size)
		return(memPtr[ofs]);
	else
	{
		Panic("Illegal memory access in RamSpace::MemRead()");
		return(MAXWORDVAL);
	}
}


// This method allows to write data to a specified address (as word offset)
// (SystemBus must check address validity and make byte-to-word
// address conversion)
void RamSpace::MemWrite(Word ofs, Word data)
{
	if(ofs < size)
		memPtr[ofs] = data;
	else
		Panic("Illegal memory access in RamSpace::MemWrite()");
}


// This method returns RamSpace size in bytes
Word RamSpace::Size(void)
{
	return(size * WORDLEN);
}

/****************************************************************************/

// This class implements ROM devices. The BIOS or Bootstrap ROM is read
// from a disk file, whose name is specified during object creation.
// Access is read-only, to word-sized items in it. SystemBus must do all
// bounds checking and address conversion for access

class BiosSpace
{
	public:
	
		// This method creates a BiosSpace object, filling with .rom file
		// contents
		BiosSpace(const char * name);

		// This method deletes the contents of a BiosSpace object
		~BiosSpace(void);
		
		// This method returns the value of Word at ofs address
		// (SystemBus must assure that ofs is in range)
		Word MemRead(Word ofs);

		// This method returns BiosSpace size in bytes
		Word Size(void);
		
	private:

		// pointer to memory structure
		Word * memPtr;
		
		// size of structure in Words (C style addressing: [0..size - 1])
		Word size;
};


// This method creates a BiosSpace object, filling with .rom file contents
BiosSpace::BiosSpace(const char * name)
{
	memPtr = readBIOSFile(name, &size);
}


// This method deletes the contents of a BiosSpace object
BiosSpace::~BiosSpace(void)
{
	delete memPtr;
}


// This method returns the value of Word at ofs address
// (SystemBus must assure that ofs is in range)
Word BiosSpace::MemRead(Word ofs)
{
	if (ofs < size)
		return(memPtr[ofs]);
	else
	{
		Panic("Illegal memory access in BiosSpace::MemRead()");
		return(MAXWORDVAL);
	}
}


// This method returns BiosSpace size in bytes
Word BiosSpace::Size(void)
{
	return(size * WORDLEN);
}


/****************************************************************************/
/* Definitions strictly local to the module.                                */
/****************************************************************************/

// This function reads the ROM contents from disk and returns a pointer to the
// memory structure created and its length (via siz pointer); it recognizes
// the file using the magic file number for .rom files.
// If no correct load is possible, ROM itself is filled with NOPs

HIDDEN Word * readBIOSFile(const char * fName, Word * sizep)
{
	FILE * bFile = NULL;
	Word tag;
	Word * biosBuf;
	
	if (fName == NULL || SAMESTRING(fName, EMPTYSTR) || (bFile = fopen(fName, "r")) == NULL || \
		fread((void *) &tag, WORDLEN, 1, bFile) != 1 || \
		tag != BIOSFILEID || fread((void *) sizep, WORDLEN, 1, bFile) != 1)
	{
		// missing type or wrong type
		biosBuf = emptyFrameSignal(fName, sizep);
	}
	else
	{
		biosBuf = new Word[*sizep];
		if (fread((void *) biosBuf, WORDLEN, *sizep, bFile) != *sizep)
		{
			// wrong file format
			delete biosBuf;
			biosBuf = emptyFrameSignal(fName, sizep);
		}
		//else all is ok
	}
	if (bFile != NULL)
		fclose(bFile);

	return(biosBuf);
}	


// This function creates a "dummy" ROM memory space filled with NOPs and
// warns user
HIDDEN Word * emptyFrameSignal(const char * fName, Word * sizep)
{
	Word * bBuf;
	unsigned int i;
	
	ShowAlert("Unable to load critical BIOS file:", fName, "replacing it with a full frame of (nop)");
	bBuf = new Word[FRAMESIZE];
	for (i = 0; i < FRAMESIZE; i++)
		bBuf[i] = NOP;
		
	*sizep = FRAMESIZE;
	return(bBuf);
}
