/* File: $Id: systembus.cc,v 1.19 1998/02/25 16:17:56 morsiani Exp $ */

/****************************************************************************
 *
 * This module implements the SystemBus object class.
 * 
 * A SystemBus connects a Processor object to memory and devices. 
 * It satisfies the Processor requests to read and write memory positions,
 * and simulates interrupts generation and other external exceptions.
 * SystemBus defines the physical memory map of the system, mapping RamSpace
 * and BiosSpace objects and Device registers to "physical" addresses in
 * memory.  
 * It also implements the system clock and timer, and keeps track of some
 * important system constants.  All this information is mapped in the "bus
 * device register area", at the start of the device register area.
 *
 * On creation, it initializes the device and memory subsystems
 * as set in the simulator configuration (contained in the SetupInfo 
 * object); on destruction, it removes them.  
 * 
 * SystemBus notifies all changes to memory location to Watch controlling
 * object, to track memory accesses for simulation brkpt/susp/trace handling
 * 
 ****************************************************************************/


/****************************************************************************/
/* Inclusion of header files.                                               */
/****************************************************************************/

#include <stdio.h>

#include "h/const.h"
#include "h/types.h"
#include "h/blockdev.h"


/****************************************************************************/
/* Inclusion of imported declarations.                                      */
/****************************************************************************/

#include "e/utility.e"
#include "e/bus.support.e"
#include "e/device.e"
#include "e/blockdev.e"
#include "e/processor.e"
#include "e/memspace.e"
#include "e/setup.e"
#include "e/watch.e"


/****************************************************************************/
/* Declarations strictly local to the module.                               */
/****************************************************************************/


// This macro converts a byte address into a word address (minus offset)
#define CONVERT(ad, bs)	((ad - bs) >> WORDSHIFT)	


/****************************************************************************/
/* Definitions to be exported.                                              */
/****************************************************************************/

class SystemBus
{
	public:
	
		// This method builds a SystemBus object and all related structures
		SystemBus(SetupInfo * stp, Watch * wch);
		// This method deletes a SystemBus object and all related structures
		~SystemBus(void);
		
		// This method links SystemBus to related Processor: it may be used
		// to build multi-processor systems with shared buses
		void LinkToProc(Processor * prc);		

		// This method increments system clock and decrements interval
		// timer; on timer underflow (0 -> FFFFFFFF transition) a interrupt
		// is generated.  Event queue is checked against the current clock
		// value and device operations are completed if needed; all memory
		// changes are notified to Watch control object
		void ClockTick(void);

		// This method reads a data word from memory at physical address
		// addr, returning it thru datap pointer. It also returns TRUE if
		// the address was invalid and an exception was caused, FALSE
		// otherwise, and signals memory access to Watch control object
		Boolean DataRead(Word addr, Word * datap);
		
		// This method writes the data word at physical addr in RAM memory or device
		// register area.  Writes to BIOS or BOOT areas cause a DBEXCEPTION (no
		// writes allowed). It returns TRUE if an exception was caused, FALSE
		// otherwise, and notifies access to Watch control object
		Boolean DataWrite(Word addr, Word data);	
		
		// This method reads a istruction from memory at physical address addr,
		// returning it thru istrp pointer. It also returns TRUE if the
		// address was invalid and an exception was caused, FALSE otherwise,
		// and notifies Watch
		Boolean InstrRead(Word addr, Word * instp);
		
		// This method transfers a block from or to memory, starting with
		// address startAddr; it returns TRUE is transfer was not successful
		// (non-existent memory, read-only memory, unaligned addresses),
		// FALSE otherwise.  It notifies too the memory accesses to Watch
		// control object
		Boolean DMATransfer(Block * blk, Word startAddr, Boolean toMemory);
		
		// This method inserts in the eventQ a event that must happen
		// at (current system time) + (increment)
		TimeStamp * EventReq(unsigned int intNum, unsigned int devNum, Word inc);
		
		// This method sets the appropriate bits into intCauseDev[] and
		// IntPendMask to signal device interrupt pending; it notifies
		// memory changes to Watch too
		void IntReq(unsigned int intNum, unsigned int devNum);

		// This method resets the appropriate bits into intCauseDev[] and
		// IntPendMask to signal device interrupt acknowlege; it notifies
		// memory changes to Watch too
		void IntAck(unsigned int intNum, unsigned int devNum);

		// This method returns the current interrupt line status
		Word getPendingInt(void);
		
		// This method returns the Device object with given "coordinates"
		Device * getDev(unsigned int intL, unsigned int dNum);
		
		
		// These methods allow to inspect or modify  TimeofDay Clock and
		// Interval Timer (typically for simulation reasons)
		
		Word getToDHI(void);
		Word getToDLO(void);
		Word getTimer(void);
		void setToDHI(Word hi);
		void setToDLO(Word lo);
		void setTimer(Word time);
		
		// These methods allow Watch to inspect or modify single memory
		// locations; they return TRUE if address is invalid or cannot be
		// changed, and FALSE otherwise
		 
		Boolean WatchRead(Word addr, Word * datap);
		Boolean WatchWrite(Word addr, Word data);

	private:

		// configuration information object
		SetupInfo * setup;
		
		// watch interface object
		Watch * watch;
		
		// Processor object
		Processor * proc;

		// system clock & interval timer
		TimeStamp * timeOfDay;
		Word timer;
		
		// device events queue
		EventQueue * eventQ;
		
		// physical memory spaces
		RamSpace * ram;
		BiosSpace * bios;
		BiosSpace * boot;
		
		// device handling & interrupt generation tables
		Device * devTable[DEVINTUSED][DEVPERINT];
		Word instDevTable[DEVINTUSED];
		Word intCauseDev[DEVINTUSED];
		
		// pending interrupts on lines: this word is packed into MIPS Cause
		// Register IP field format for easy masking
		Word intPendMask;
		
		// This method read the data at physical address addr, and
		// passes it back thru the datap pointer. It also return FALSE if
		// the addr is valid, and TRUE otherwise
		Boolean busRead(Word addr, Word * datap);
		
		// This method returns the value for the device field addressed in
		// the "bus register area"
		Word busRegRead(Word addr);
		
		// This method writes the data at physical address addr, and
		// passes it back thru the datap pointer. It also return FALSE if
		// the addr is valid and writable, and TRUE otherwise
		Boolean busWrite(Word addr, Word data);
		
		// This method accesses the system configuration and constructs
		// the devices needed, linking them to SystemBus object
		Device * makeDev(unsigned int intl, unsigned int dnum);	
};


// This method builds a SystemBus object and all related structures
SystemBus::SystemBus(SetupInfo * stp, Watch * wch)
{
	unsigned int intl, dnum;
	
	setup = stp;
	watch = wch;
	proc = NULL;

	// init clock, interval timer and device events queue
	timeOfDay = new TimeStamp(0UL, 0UL);
	timer = MAXWORDVAL;
	
	eventQ = new EventQueue();
	
	// create RAM and ROM spaces
	
	if (setup->getBootType() == COREBOOT)
		ram = new RamSpace(setup->getRamSize() * FRAMESIZE, setup->getROMFileName(COREINDEX));
	else
		// does not initialize ram space with core file, even if it is defined
		ram = new RamSpace(setup->getRamSize() * FRAMESIZE, NULL);
	
	// creates and loads ROMs	
	bios = new BiosSpace(setup->getROMFileName(BIOSINDEX));
	boot = new BiosSpace(setup->getROMFileName(BOOTINDEX));
	
	// create devices and initialize registers used for interrupt handling 
	intPendMask = 0UL;
	for (intl = 0; intl < DEVINTUSED; intl++)
	{
		instDevTable[intl] = 0UL;
		intCauseDev[intl] = 0UL;
		for (dnum = 0; dnum < DEVPERINT; dnum++)
		{
			devTable[intl][dnum] = makeDev(intl, dnum);
			if (devTable[intl][dnum]->Type() != NULLDEV)
				// device is installed
				instDevTable[intl] = SetBit(instDevTable[intl], dnum);
		} 	
	}
}


// This method deletes a SystemBus object and all related structures
SystemBus::~SystemBus(void)
{
	unsigned int intl, dnum;
	
	delete timeOfDay;
	delete eventQ;
	
	delete ram;
	delete bios;
	delete boot;
	
	for (intl = 0; intl < DEVINTUSED; intl++)
		for (dnum = 0; dnum < DEVPERINT; dnum++)
			delete devTable[intl][dnum];
}


// This method links SystemBus to related Processor: it may be used to build 
// multi-processor systems with shared buses
void SystemBus::LinkToProc(Processor * prc)
{
	proc = prc;
}


// This method increments system clock and decrements interval timer;
// on timer underflow (0 -> FFFFFFFF transition) a interrupt is generated.
// Event queue is checked against the current clock value and device operations
// are completed if needed; all memory changes are notified to Watch control
// object
void SystemBus::ClockTick(void)
{
	unsigned int devAddr;
	unsigned int intl, dnum;
	
	timeOfDay->Increase();
	
	// both registers signal "change" because they are conceptually one
	watch->SignalBusAccess((TODHIADDR * WORDLEN) + DEVBASE, WRITE);
	watch->SignalBusAccess((TODLOADDR * WORDLEN) + DEVBASE, WRITE);
	
	if (UnsSub(&timer, timer, 1))
		// timer interrupt is now pending
		intPendMask = SetBit(intPendMask, TIMERINT + IPMASKBASE);
	
	watch->SignalBusAccess((TIMERADDR * WORDLEN) + DEVBASE, WRITE);

	// scans the event queue	
	while (!eventQ->IsEmptyQ() && (eventQ->getHTS())->LessEq(timeOfDay))
	{
		// a device operation need to be completed and that event removed
		intl = eventQ->getHIntLine();
		dnum = eventQ->getHDevNum();
		devAddr = (devTable[intl][dnum])->CompleteDevOp();
		eventQ->RemoveHead();

		// a device op completion changes device registers: memory is altered
		watch->SignalBusAccess((((DEVREGLEN * DEVPERINT * intl) + (dnum * DEVREGLEN) + devAddr) * WORDLEN) + DEVSTART, WRITE);				
	}
	
}


//
// The following methods allow to inspect or modify  TimeofDay Clock and
// Interval Timer (typically for simulation reasons); they are self-explanatory
//

Word SystemBus::getToDHI(void)
{
	return(timeOfDay->getHiTS());
}

Word SystemBus::getToDLO(void)
{
	return(timeOfDay->getLoTS());
}

Word SystemBus::getTimer(void)
{
	return(timer);
}
void SystemBus::setToDHI(Word hi)
{
	timeOfDay->setHiTS(hi);
}
void SystemBus::setToDLO(Word lo)
{
	timeOfDay->setLoTS(lo);
}

void SystemBus::setTimer(Word time)
{
	timer = time;
}


// This method reads a data word from memory at address addr, returning it
// thru datap pointer. It also returns TRUE if the address was invalid and
// an exception was caused, FALSE otherwise, and signals memory access to
// Watch control object
Boolean SystemBus::DataRead(Word addr, Word * datap)
{
	watch->SignalBusAccess(addr, READ);
	
	if (busRead(addr, datap))
	{
		// address invalid: signal exception to processor
		proc->SignalExc(DBEXCEPTION);
		return(TRUE);
	}
	else
		// address was valid
		return(FALSE);
		
}


//
// These methods allow Watch to inspect or modify single memory locations;
// they return TRUE if address is invalid or memory cannot be altered, and
// FALSE otherwise
// 

Boolean SystemBus::WatchRead(Word addr, Word * datap)
{
	return(busRead(addr, datap));
}

Boolean SystemBus::WatchWrite(Word addr, Word data)
{	
	return(busWrite(addr, data));
}


// This method writes the data word at physical addr in RAM memory or device
// register area.  Writes to BIOS or BOOT areas cause a DBEXCEPTION (no
// writes allowed). It returns TRUE if an exception was caused, FALSE
// otherwise, and notifies access to Watch control object
Boolean SystemBus::DataWrite(Word addr, Word data)
{
	watch->SignalBusAccess(addr, WRITE);

	if (busWrite(addr, data))
	{
		// data write is out of valid write bounds
		proc->SignalExc(DBEXCEPTION);
		return(TRUE);
	}
	else
		// write was successful
		return(FALSE);
}


// This method transfers a block from or to memory, starting with address
// startAddr; it returns TRUE is transfer was not successful (non-existent
// memory, read-only memory, unaligned addresses), FALSE otherwise.
// It notifies too the memory accesses to Watch control object
Boolean SystemBus::DMATransfer(Block * blk, Word startAddr, Boolean toMemory)
{
	Word ofs, val;
	Boolean error = FALSE;
	
	if (BADADDR(startAddr))
		error = TRUE;
	else
		if (toMemory)
			for (ofs = 0; ofs < BLOCKSIZE && !error; ofs++)
			{
				error = busWrite(startAddr + (ofs * WORDLEN), blk->getWord(ofs));  
				watch->SignalBusAccess(startAddr + (ofs * WORDLEN), WRITE);
			}
		else
			for (ofs = 0; ofs < BLOCKSIZE && !error; ofs++)
			{
				error = busRead(startAddr + (ofs * WORDLEN), &val);
				watch->SignalBusAccess(startAddr + (ofs * WORDLEN), READ);
				blk->setWord(ofs, val);
			}
	return (error);
}

				
// This method reads a istruction from memory at address addr, returning
// it thru istrp pointer. It also returns TRUE if the address was invalid and
// an exception was caused, FALSE otherwise, and notifies Watch
Boolean SystemBus::InstrRead(Word addr, Word * instrp)
{
	watch->SignalBusAccess(addr, EXEC);
	
	if (busRead(addr, instrp))
	{
		// address invalid: signal exception to processor
		proc->SignalExc(IBEXCEPTION);
		return(TRUE);
	}
	else
		// address was valid
		return(FALSE);
		
}


// This method inserts in the eventQ a event that must happen
// at (current system time) + (increment)
TimeStamp * SystemBus::EventReq(unsigned int intNum, unsigned int devNum, Word inc)
{
	return(eventQ->InsertQ(timeOfDay, inc, intNum, devNum));
}


// This method sets the appropriate bits into intCauseDev[] and IntPendMask
// to signal device interrupt pending; it notifies memory changes to Watch
// too
void SystemBus::IntReq(unsigned int intl, unsigned int devNum)
{
	intCauseDev[intl] = SetBit(intCauseDev[intl], devNum);
	intPendMask = SetBit(intPendMask, intl + DEVINTBASE + IPMASKBASE);
	// memory address changes
	watch->SignalBusAccess(((CDEVBASEAD + intl) * WORDLEN) + DEVBASE, WRITE);
}


// This method resets the appropriate bits into intCauseDev[] and
// IntPendMask to signal device interrupt acknowlege; it notifies memory
// changes to Watch too
void SystemBus::IntAck(unsigned int intl, unsigned int devNum)
{
	intCauseDev[intl] = ResetBit(intCauseDev[intl], devNum);
	// memory address changes
	watch->SignalBusAccess(((CDEVBASEAD + intl) * WORDLEN) + DEVBASE, WRITE);
	if (intCauseDev[intl] == 0UL)
		// there are no more interrupts pending for that line
		intPendMask = ResetBit(intPendMask, intl + DEVINTBASE + IPMASKBASE);
}


// This method returns the current interrupt line status
Word SystemBus::getPendingInt(void)
{
	return(intPendMask);
}


// This method returns the Device object with given "coordinates"
Device * SystemBus::getDev(unsigned int intL, unsigned int dNum)
{
	if (intL < DEVINTUSED  && dNum < DEVPERINT)
		return(devTable[intL][dNum]);
	else
	{
		Panic("Unknown device specified in SystemBus::getDev()");
		// never returns
		return(NULL);
	}
}


/****************************************************************************/
/* Definitions strictly local to the module.                                */
/****************************************************************************/

// This method reads the data at the address addr, and passes it back thru
// the datap pointer. It also return FALSE if the addr is valid, and TRUE
// otherwise
Boolean SystemBus::busRead(Word addr, Word * datap)
{
	if (INBOUNDS(addr, RAMBASE, RAMBASE + ram->Size()))
		*datap = ram->MemRead(CONVERT(addr, RAMBASE));
	else
		if (INBOUNDS(addr, BIOSBASE, BIOSBASE + bios->Size()))
			*datap = bios->MemRead(CONVERT(addr,BIOSBASE));
		else
			if (INBOUNDS(addr, BOOTBASE, BOOTBASE + boot->Size()))
				*datap = boot->MemRead(CONVERT(addr, BOOTBASE));
			else
				if (INBOUNDS(addr, DEVBASE, DEVEND))
					*datap = busRegRead(addr);
				else
				{
					// address invalid: data read is out of bounds
					*datap = MAXWORDVAL;
					return(TRUE);
				}
	// address was valid 
	return(FALSE);
}


// This method returns the value for the device field addressed in the "bus
// register area"
Word SystemBus::busRegRead(Word addr)
{
	unsigned int intl, dnum;
	Word data;
	
	// compute offset word count
	addr = CONVERT(addr, DEVBASE);

	if (addr >= DEVREGBADR)
	{
		addr -= DEVREGBADR;
		// data read is in device registers area
		intl = addr / (DEVPERINT * DEVREGLEN);
		dnum = (addr / DEVREGLEN) % DEVPERINT;
		data = (devTable[intl][dnum])->ReadDevReg(addr % DEVREGLEN);
	}
	else
		if (INBOUNDS(addr, IDEVBASEAD, IDEVENDADR))
			// data read is in instDevTable array space
			data = instDevTable[addr - IDEVBASEAD];
		else
			if (INBOUNDS(addr, CDEVBASEAD, CDEVENDADR))
				// data read is in intCauseDev array space
				data = intCauseDev[addr - CDEVBASEAD];
			else
				// data read is in lower bus registers area 
				switch (addr)
				{
					case TIMESCALEA:
						// number of ticks for a microsecond
						data = setup->getSpeed();
						break;
					case TODHIADDR:
						data = timeOfDay->getHiTS();
						break;
					case TODLOADDR:
						data = timeOfDay->getLoTS();
						break;
					case TIMERADDR:
						data = timer;
						break;
					case RAMBASEADR:
						data = RAMBASE;
						break;
					case RAMSIZEADDR:
						data = ram->Size();
						break;
					case BIOSBASEAD:
						data = BIOSBASE;
						break;
					case BIOSSIZEADR:
						data = bios->Size();
						break;
					case BOOTBASEAD:
						data = BOOTBASE;
						break;
					case BOOTSIZEADR:
						data = boot->Size();
						break;
					default:
						// unmapped bus device register area:
						// read give 0, write has no effects
						data = 0UL;
						break;
				}
	return(data);
}


// This method accesses the system configuration and constructs
// the devices needed, linking them to SystemBus object
Device * SystemBus::makeDev(unsigned int intl, unsigned int dnum)
{
	unsigned int devt;
	Device * dev;
	
	devt = setup->getDevType(intl, dnum);
	switch(devt)
	{

		case PRNTDEV:
			dev = new PrinterDevice(this, setup, intl, dnum);
			break;
		
		case TERMDEV:
			dev = new TerminalDevice(this, setup, intl, dnum);
			break;
			
		case DISKDEV:
			dev = new DiskDevice(this, setup, intl, dnum);
			break;
		
		case TAPEDEV:
			dev = new TapeDevice(this, setup, intl, dnum);
			break;
			
		default:
			dev = new Device(this, intl, dnum);
			break;
	}
	return(dev);
}

// This method writes the data at the physical address addr, and passes it
// back thru the datap pointer. It also return FALSE if the addr is valid
// and writable, and TRUE otherwise
Boolean SystemBus::busWrite(Word addr, Word data)
{
	unsigned int intl, dnum;
	
	if (INBOUNDS(addr, RAMBASE, RAMBASE + ram->Size()))
	{
		ram->MemWrite(CONVERT(addr, RAMBASE), data);
	}
	else 
		if (INBOUNDS(addr, DEVBASE, DEVEND))
		{
			// data write is in bus & device registers area
			if (addr >= DEVSTART)
			{
				// data write is on a device register: let's compute which
				addr = CONVERT(addr, DEVSTART);
				intl = addr / (DEVPERINT * DEVREGLEN);
				dnum = (addr / DEVREGLEN) % DEVPERINT;
				(devTable[intl][dnum])->WriteDevReg(addr % DEVREGLEN, data);
			}
			else
				// data write is in bus registers area
				if (addr == DEVBASE + (TIMERADDR * WORDLEN))
				{
					// update the interval timer and reset its interrupt line
					timer = data;
					intPendMask = ResetBit(intPendMask, TIMERINT + IPMASKBASE);
				}	
				// else data write is on a read only bus register, and
				// has no harmful effects
		}
		else
			// data write out from valid write bounds
			return(TRUE);

	// into any valid case		
	return(FALSE);
}
